@@ -23,15 +23,15 @@
  * or implied warranty.
  */
 #include "k5-int.h"
+#include "k5-der.h"
 #include "gssapiP_krb5.h"
 
 /*
  * IAKERB implementation
  */
 
-extern int gssint_get_der_length(unsigned char **, OM_uint32, unsigned int*);
-
 enum iakerb_state {
+    IAKERB_REALM_DISCOVERY, /* querying server for its realm */
     IAKERB_AS_REQ,      /* acquiring ticket with initial creds */
     IAKERB_TGS_REQ,     /* acquiring ticket with TGT */
     IAKERB_AP_REQ       /* hand-off to normal GSS AP-REQ exchange */
@@ -78,6 +78,46 @@ iakerb_release_context(iakerb_ctx_id_t ctx)
     free(ctx);
 }
 
+/* Encode a KRB-ERROR message with the given protocol code.  Use the server
+ * principal from verifier_cred if one is available. */
+static krb5_error_code
+iakerb_mk_error(krb5_context context, gss_cred_id_t verifier_cred,
+                int protocol_err, krb5_data *enc_err)
+{
+    krb5_error error = { 0 };
+    krb5_gss_cred_id_t cred = (krb5_gss_cred_id_t)verifier_cred;
+
+    error.error = protocol_err;
+
+    /* We must provide a server principal, although we expect the recipient to
+     * care chiefly about the error code. */
+    if (cred != NULL && cred->name != NULL)
+        error.server = cred->name->princ;
+    else
+        error.server = (krb5_principal)krb5_anonymous_principal();
+
+    return krb5_mk_error(context, &error, enc_err);
+}
+
+/* Decode a KRB-ERROR message and return the associated com_err code. */
+static krb5_error_code
+iakerb_rd_error(krb5_context context, const krb5_data *enc_err)
+{
+    krb5_error_code ret;
+    krb5_error *error;
+
+    ret = krb5_rd_error(context, enc_err, &error);
+    if (ret)
+        return ret;
+
+    if (error->error > 0 && error->error <= KRB_ERR_MAX)
+        ret = error->error + ERROR_TABLE_BASE_krb5;
+    else
+        ret = KRB5KRB_ERR_GENERIC;
+    krb5_free_error(context, error);
+    return ret;
+}
+
 /*
  * Create a IAKERB-FINISHED structure containing a checksum of
  * the entire IAKERB exchange.
@@ -98,8 +138,8 @@ iakerb_make_finished(krb5_context context,
     if (key == NULL)
         return KRB5KDC_ERR_NULL_KEY;
 
-    code = krb5_k_make_checksum(context, 0, key, KRB5_KEYUSAGE_IAKERB_FINISHED,
-                                conv, &iaf.checksum);
+    code = krb5_k_make_checksum(context, 0, key, KRB5_KEYUSAGE_FINISHED, conv,
+                                &iaf.checksum);
     if (code != 0)
         return code;
 
@@ -130,8 +170,8 @@ iakerb_verify_finished(krb5_context context,
     if (code != 0)
         return code;
 
-    code = krb5_k_verify_checksum(context, key, KRB5_KEYUSAGE_IAKERB_FINISHED,
-                                  conv, &iaf->checksum, &valid);
+    code = krb5_k_verify_checksum(context, key, KRB5_KEYUSAGE_FINISHED, conv,
+                                  &iaf->checksum, &valid);
     if (code == 0 && valid == FALSE)
         code = KRB5KRB_AP_ERR_BAD_INTEGRITY;
 
@@ -164,7 +204,6 @@ iakerb_save_token(iakerb_ctx_id_t ctx, const gss_buffer_t token)
  */
 static krb5_error_code
 iakerb_parse_token(iakerb_ctx_id_t ctx,
-                   int initialContextToken,
                    const gss_buffer_t token,
                    krb5_data *realm,
                    krb5_data **cookie,
@@ -172,46 +211,30 @@ iakerb_parse_token(iakerb_ctx_id_t ctx,
 {
     krb5_error_code code;
     krb5_iakerb_header *iah = NULL;
-    unsigned int bodysize, lenlen;
-    int length;
-    unsigned char *ptr;
-    int flags = 0;
+    const uint8_t *token_body;
     krb5_data data;
+    struct k5input in, seq;
 
     if (token == GSS_C_NO_BUFFER || token->length == 0) {
         code = KRB5_BAD_MSIZE;
         goto cleanup;
     }
 
-    if (initialContextToken)
-        flags |= G_VFY_TOKEN_HDR_WRAPPER_REQUIRED;
-
-    ptr = token->value;
-
-    code = g_verify_token_header(gss_mech_iakerb,
-                                 &bodysize, &ptr,
-                                 IAKERB_TOK_PROXY,
-                                 token->length, flags);
-    if (code != 0)
+    k5_input_init(&in, token->value, token->length);
+    if (!g_verify_token_header(&in, gss_mech_iakerb) ||
+        k5_input_get_uint16_be(&in) != IAKERB_TOK_PROXY) {
+        code = G_BAD_TOK_HEADER;
         goto cleanup;
+    }
 
-    data.data = (char *)ptr;
-
-    if (bodysize-- == 0 || *ptr++ != 0x30 /* SEQUENCE */) {
+    /* Find the end of the DER sequence tag and decode it (with the tag) as the
+     * IAKERB header. */
+    token_body = in.ptr;
+    if (!k5_der_get_value(&in, 0x30, &seq)) {
         code = ASN1_BAD_ID;
         goto cleanup;
     }
-
-    length = gssint_get_der_length(&ptr, bodysize, &lenlen);
-    if (length < 0 || bodysize - lenlen < (unsigned int)length) {
-        code = KRB5_BAD_MSIZE;
-        goto cleanup;
-    }
-    data.length = 1 /* SEQUENCE */ + lenlen + length;
-
-    ptr += length;
-    bodysize -= (lenlen + length);
-
+    data = make_data((uint8_t *)token_body, seq.ptr + seq.len - token_body);
     code = decode_krb5_iakerb_header(&data, &iah);
     if (code != 0)
         goto cleanup;
@@ -226,9 +249,8 @@ iakerb_parse_token(iakerb_ctx_id_t ctx,
         iah->cookie = NULL;
     }
 
-    request->data = (char *)ptr;
-    request->length = bodysize;
-
+    /* The remainder of the token body is the request. */
+    *request = make_data((uint8_t *)in.ptr, in.len);
     assert(request->data + request->length ==
            (char *)token->value + token->length);
 
@@ -239,14 +261,14 @@ cleanup:
 }
 
 /*
- * Create a token from IAKERB-HEADER and KRB-KDC-REQ/REP
+ * Create a token from IAKERB-HEADER and KRB-KDC-REQ/REP.  Save the generated
+ * token for the finish checksum and increment the message count.
  */
 static krb5_error_code
 iakerb_make_token(iakerb_ctx_id_t ctx,
                   krb5_data *realm,
                   krb5_data *cookie,
                   krb5_data *request,
-                  int initialContextToken,
                   gss_buffer_t token)
 {
     krb5_error_code code;
@@ -254,7 +276,7 @@ iakerb_make_token(iakerb_ctx_id_t ctx,
     krb5_data *data = NULL;
     char *p;
     unsigned int tokenSize;
-    unsigned char *q;
+    struct k5buf buf;
 
     token->value = NULL;
     token->length = 0;
@@ -283,29 +305,23 @@ iakerb_make_token(iakerb_ctx_id_t ctx,
         memcpy(data->data + data->length, request->data, request->length);
     data->length += request->length;
 
-    if (initialContextToken)
-        tokenSize = g_token_size(gss_mech_iakerb, data->length);
-    else
-        tokenSize = 2 + data->length;
-
-    token->value = q = gssalloc_malloc(tokenSize);
-    if (q == NULL) {
+    tokenSize = g_token_size(gss_mech_iakerb, data->length);
+    token->value = gssalloc_malloc(tokenSize);
+    if (token->value == NULL) {
         code = ENOMEM;
         goto cleanup;
     }
     token->length = tokenSize;
+    k5_buf_init_fixed(&buf, token->value, token->length);
 
-    if (initialContextToken) {
-        g_make_token_header(gss_mech_iakerb, data->length, &q,
-                            IAKERB_TOK_PROXY);
-    } else {
-        store_16_be(IAKERB_TOK_PROXY, q);
-        q += 2;
-    }
-    memcpy(q, data->data, data->length);
-    q += data->length;
+    g_make_token_header(&buf, gss_mech_iakerb, data->length, IAKERB_TOK_PROXY);
+    k5_buf_add_len(&buf, data->data, data->length);
+    assert(buf.len == token->length);
 
-    assert(q == (unsigned char *)token->value + token->length);
+    code = iakerb_save_token(ctx, token);
+    if (code != 0)
+        goto cleanup;
+    ctx->count++;
 
 cleanup:
     krb5_free_data(ctx->k5c, data);
@@ -313,6 +329,47 @@ cleanup:
     return code;
 }
 
+/* Generate a response to a realm discovery request. */
+static krb5_error_code
+iakerb_acceptor_realm(iakerb_ctx_id_t ctx, gss_cred_id_t verifier_cred,
+                      gss_buffer_t output_token)
+{
+    krb5_error_code ret;
+    OM_uint32 dummy;
+    krb5_gss_cred_id_t cred = (krb5_gss_cred_id_t)verifier_cred;
+    krb5_data realm = empty_data(), reply = empty_data();
+    char *defrealm = NULL;
+
+    /* Get the acceptor realm from the verifier cred if we can; otherwise try
+     * to use the default realm. */
+    if (cred != NULL && cred->name != NULL &&
+        cred->name->princ->realm.length > 0) {
+        realm = cred->name->princ->realm;
+    } else {
+        ret = krb5_get_default_realm(ctx->k5c, &defrealm);
+        if (ret) {
+            /* Generate an error reply if there is no default realm. */
+            ret = iakerb_mk_error(ctx->k5c, verifier_cred,
+                                  KRB_AP_ERR_IAKERB_KDC_NOT_FOUND, &reply);
+            if (ret)
+                goto cleanup;
+        } else {
+            realm = string2data(defrealm);
+        }
+    }
+
+    ret = iakerb_make_token(ctx, &realm, NULL, &reply, output_token);
+    if (ret)
+        goto cleanup;
+
+cleanup:
+    if (ret)
+        gss_release_buffer(&dummy, output_token);
+    krb5_free_default_realm(ctx->k5c, defrealm);
+    krb5_free_data_contents(ctx->k5c, &reply);
+    return ret;
+}
+
 /*
  * Parse the IAKERB token in input_token and send the contained KDC
  * request to the KDC for the realm.
@@ -320,8 +377,7 @@ cleanup:
  * Wrap the KDC reply in output_token.
  */
 static krb5_error_code
-iakerb_acceptor_step(iakerb_ctx_id_t ctx,
-                     int initialContextToken,
+iakerb_acceptor_step(iakerb_ctx_id_t ctx, gss_cred_id_t verifier_cred,
                      const gss_buffer_t input_token,
                      gss_buffer_t output_token)
 {
@@ -329,7 +385,7 @@ iakerb_acceptor_step(iakerb_ctx_id_t ctx,
     krb5_data request = empty_data(), reply = empty_data();
     krb5_data realm = empty_data();
     OM_uint32 tmp;
-    int tcp_only, use_primary;
+    int tcp_only, use_primary, protocol_err;
     krb5_ui_4 kdc_code;
 
     output_token->length = 0;
@@ -340,20 +396,23 @@ iakerb_acceptor_step(iakerb_ctx_id_t ctx,
         goto cleanup;
     }
 
-    code = iakerb_parse_token(ctx, initialContextToken, input_token, &realm,
-                              NULL, &request);
+    code = iakerb_parse_token(ctx, input_token, &realm, NULL, &request);
     if (code != 0)
         goto cleanup;
 
-    if (realm.length == 0 || request.length == 0) {
-        code = KRB5_BAD_MSIZE;
-        goto cleanup;
-    }
-
     code = iakerb_save_token(ctx, input_token);
     if (code != 0)
         goto cleanup;
 
+    if (realm.length == 0 && request.length == 0) {
+        /* This is a realm discovery request. */
+        code = iakerb_acceptor_realm(ctx, verifier_cred, output_token);
+        goto cleanup;
+    } else if (realm.length == 0 || request.length == 0) {
+        code = KRB5_BAD_MSIZE;
+        goto cleanup;
+    }
+
     for (tcp_only = 0; tcp_only <= 1; tcp_only++) {
         use_primary = 0;
         code = krb5_sendto_kdc(ctx->k5c, &request, &realm,
@@ -376,29 +435,16 @@ iakerb_acceptor_step(iakerb_ctx_id_t ctx,
     }
 
     if (code == KRB5_KDC_UNREACH || code == KRB5_REALM_UNKNOWN) {
-        krb5_error error;
-
-        memset(&error, 0, sizeof(error));
-        if (code == KRB5_KDC_UNREACH)
-            error.error = KRB_AP_ERR_IAKERB_KDC_NO_RESPONSE;
-        else if (code == KRB5_REALM_UNKNOWN)
-            error.error = KRB_AP_ERR_IAKERB_KDC_NOT_FOUND;
-
-        code = krb5_mk_error(ctx->k5c, &error, &reply);
+        protocol_err = (code == KRB5_KDC_UNREACH) ?
+            KRB_AP_ERR_IAKERB_KDC_NO_RESPONSE :
+            KRB_AP_ERR_IAKERB_KDC_NOT_FOUND;
+        code = iakerb_mk_error(ctx->k5c, verifier_cred, protocol_err, &reply);
         if (code != 0)
             goto cleanup;
     } else if (code != 0)
         goto cleanup;
 
-    code = iakerb_make_token(ctx, &realm, NULL, &reply, 0, output_token);
-    if (code != 0)
-        goto cleanup;
-
-    code = iakerb_save_token(ctx, output_token);
-    if (code != 0)
-        goto cleanup;
-
-    ctx->count++;
+    code = iakerb_make_token(ctx, &realm, NULL, &reply, output_token);
 
 cleanup:
     if (code != 0)
@@ -530,26 +576,63 @@ iakerb_initiator_step(iakerb_ctx_id_t ctx,
                       gss_buffer_t output_token)
 {
     krb5_error_code code = 0;
-    krb5_data in = empty_data(), out = empty_data(), realm = empty_data();
+    krb5_data in = empty_data(), out = empty_data();
+    krb5_data realm = empty_data(), server_realm = empty_data();
     krb5_data *cookie = NULL;
     OM_uint32 tmp;
     unsigned int flags = 0;
     krb5_ticket_times times;
+    krb5_boolean first_token;
 
     output_token->length = 0;
     output_token->value = NULL;
 
-    if (input_token != GSS_C_NO_BUFFER) {
-        code = iakerb_parse_token(ctx, 0, input_token, NULL, &cookie, &in);
+    first_token = (input_token == GSS_C_NO_BUFFER || input_token->length == 0);
+    if (!first_token) {
+        code = iakerb_parse_token(ctx, input_token, &server_realm, &cookie,
+                                  &in);
         if (code != 0)
             goto cleanup;
 
         code = iakerb_save_token(ctx, input_token);
         if (code != 0)
             goto cleanup;
+
+        if (krb5_is_krb_error(&in)) {
+            code = iakerb_rd_error(ctx->k5c, &in);
+            if (code == KRB5KRB_AP_ERR_IAKERB_KDC_NOT_FOUND &&
+                ctx->state == IAKERB_REALM_DISCOVERY) {
+                save_error_string(code, _("The IAKERB proxy could not "
+                                          "determine its realm"));
+            }
+            if (code == KRB5KRB_AP_ERR_IAKERB_KDC_NOT_FOUND ||
+                code == KRB5KRB_AP_ERR_IAKERB_KDC_NO_RESPONSE)
+                goto cleanup;
+            code = 0;
+        }
     }
 
     switch (ctx->state) {
+    case IAKERB_REALM_DISCOVERY:
+        if (first_token) {
+            /* Send the discovery request. */
+            code = iakerb_make_token(ctx, &realm, cookie, &out, output_token);
+            goto cleanup;
+        }
+
+        /* The acceptor should have sent us its realm. */
+        if (server_realm.length == 0) {
+            code = KRB5_BAD_MSIZE;
+            goto cleanup;
+        }
+
+        /* Steal the received server realm for the client principal. */
+        krb5_free_data_contents(ctx->k5c, &cred->name->princ->realm);
+        cred->name->princ->realm = server_realm;
+        server_realm = empty_data();
+
+        ctx->state = IAKERB_AS_REQ;
+        /* Done with realm discovery; fall through to AS request. */
     case IAKERB_AS_REQ:
         if (ctx->icc == NULL) {
             code = iakerb_init_creds_ctx(ctx, cred, time_req);
@@ -608,19 +691,7 @@ iakerb_initiator_step(iakerb_ctx_id_t ctx,
 
     if (out.length != 0) {
         assert(ctx->state != IAKERB_AP_REQ);
-
-        code = iakerb_make_token(ctx, &realm, cookie, &out,
-                                 (input_token == GSS_C_NO_BUFFER),
-                                 output_token);
-        if (code != 0)
-            goto cleanup;
-
-        /* Save the token for generating a future checksum */
-        code = iakerb_save_token(ctx, output_token);
-        if (code != 0)
-            goto cleanup;
-
-        ctx->count++;
+        code = iakerb_make_token(ctx, &realm, cookie, &out, output_token);
     }
 
 cleanup:
@@ -628,6 +699,7 @@ cleanup:
         gss_release_buffer(&tmp, output_token);
     krb5_free_data(ctx->k5c, cookie);
     krb5_free_data_contents(ctx->k5c, &out);
+    krb5_free_data_contents(ctx->k5c, &server_realm);
     krb5_free_data_contents(ctx->k5c, &realm);
 
     return code;
@@ -647,6 +719,11 @@ iakerb_get_initial_state(iakerb_ctx_id_t ctx,
     krb5_creds in_creds, *out_creds = NULL;
     krb5_error_code code;
 
+    if (cred->name->princ->realm.length == 0) {
+        *state = IAKERB_REALM_DISCOVERY;
+        return 0;
+    }
+
     memset(&in_creds, 0, sizeof(in_creds));
 
     in_creds.client = cred->name->princ;
@@ -750,16 +827,11 @@ iakerb_gss_delete_sec_context(OM_uint32 *minor_status,
 static krb5_boolean
 iakerb_is_iakerb_token(const gss_buffer_t token)
 {
-    krb5_error_code code;
-    unsigned int bodysize = token->length;
-    unsigned char *ptr = token->value;
+    struct k5input in;
 
-    code = g_verify_token_header(gss_mech_iakerb,
-                                 &bodysize, &ptr,
-                                 IAKERB_TOK_PROXY,
-                                 token->length, 0);
-
-    return (code == 0);
+    k5_input_init(&in, token->value, token->length);
+    return g_verify_token_header(&in, gss_mech_iakerb) &&
+        k5_input_get_uint16_be(&in) == IAKERB_TOK_PROXY;
 }
 
 static void
@@ -787,9 +859,9 @@ iakerb_gss_accept_sec_context(OM_uint32 *minor_status,
     OM_uint32 major_status = GSS_S_FAILURE;
     OM_uint32 code;
     iakerb_ctx_id_t ctx;
-    int initialContextToken = (*context_handle == GSS_C_NO_CONTEXT);
+    krb5_boolean first_token = (*context_handle == GSS_C_NO_CONTEXT);
 
-    if (initialContextToken) {
+    if (first_token) {
         code = iakerb_alloc_context(&ctx, 0);
         if (code != 0)
             goto cleanup;
@@ -804,16 +876,12 @@ iakerb_gss_accept_sec_context(OM_uint32 *minor_status,
             major_status = GSS_S_DEFECTIVE_TOKEN;
             goto cleanup;
         }
-        code = iakerb_acceptor_step(ctx, initialContextToken,
-                                    input_token, output_token);
+        code = iakerb_acceptor_step(ctx, verifier_cred_handle, input_token,
+                                    output_token);
         if (code == (OM_uint32)KRB5_BAD_MSIZE)
             major_status = GSS_S_DEFECTIVE_TOKEN;
         if (code != 0)
             goto cleanup;
-        if (initialContextToken) {
-            *context_handle = (gss_ctx_id_t)ctx;
-            ctx = NULL;
-        }
         if (src_name != NULL)
             *src_name = GSS_C_NO_NAME;
         if (ret_flags != NULL)
@@ -848,9 +916,13 @@ iakerb_gss_accept_sec_context(OM_uint32 *minor_status,
         *mech_type = gss_mech_iakerb;
 
 cleanup:
-    if (initialContextToken && GSS_ERROR(major_status)) {
-        iakerb_release_context(ctx);
-        *context_handle = GSS_C_NO_CONTEXT;
+    if (first_token) {
+        if (GSS_ERROR(major_status)) {
+            iakerb_release_context(ctx);
+            *context_handle = GSS_C_NO_CONTEXT;
+        } else {
+            *context_handle = (gss_ctx_id_t)ctx;
+        }
     }
 
     *minor_status = code;
