@@ -30,13 +30,11 @@
  */
 
 #include "pkinit.h"
-#include <dlfcn.h>
-#include <dirent.h>
 
 static void
 free_list(char **list)
 {
-    int i;
+    size_t i;
 
     if (list == NULL)
         return;
@@ -49,7 +47,7 @@ free_list(char **list)
 static krb5_error_code
 copy_list(char ***dst, char **src)
 {
-    int i;
+    size_t i;
     char **newlist;
 
     if (dst == NULL)
@@ -311,17 +309,17 @@ parse_fs_options(krb5_context context,
                  const char *residual)
 {
     char *certname, *keyname, *save;
-    char *cert_filename = NULL, *key_filename = NULL;
+    char *copy = NULL, *cert_filename = NULL, *key_filename = NULL;
     krb5_error_code retval = ENOMEM;
 
     if (residual == NULL || residual[0] == '\0' || residual[0] == ',')
         return EINVAL;
 
-    certname = strdup(residual);
-    if (certname == NULL)
+    copy = strdup(residual);
+    if (copy == NULL)
         goto cleanup;
 
-    certname = strtok_r(certname, ",", &save);
+    certname = strtok_r(copy, ",", &save);
     if (certname == NULL)
         goto cleanup;
     keyname = strtok_r(NULL, ",", &save);
@@ -334,13 +332,15 @@ parse_fs_options(krb5_context context,
     if (key_filename == NULL)
         goto cleanup;
 
+    free(idopts->cert_filename);
+    free(idopts->key_filename);
     idopts->cert_filename = cert_filename;
     idopts->key_filename = key_filename;
     cert_filename = key_filename = NULL;
     retval = 0;
 
 cleanup:
-    free(certname);
+    free(copy);
     free(cert_filename);
     free(key_filename);
     return retval;
@@ -356,10 +356,12 @@ parse_pkcs12_options(krb5_context context,
     if (residual == NULL || residual[0] == '\0')
         return 0;
 
+    free(idopts->cert_filename);
     idopts->cert_filename = strdup(residual);
     if (idopts->cert_filename == NULL)
         goto cleanup;
 
+    free(idopts->key_filename);
     idopts->key_filename = strdup(residual);
     if (idopts->key_filename == NULL)
         goto cleanup;
@@ -384,8 +386,7 @@ process_option_identity(krb5_context context,
     int idtype;
     krb5_error_code retval = 0;
 
-    pkiDebug("%s: processing value '%s'\n",
-             __FUNCTION__, value ? value : "NULL");
+    TRACE_PKINIT_IDENTITY_OPTION(context, value);
     if (value == NULL)
         return EINVAL;
 
@@ -439,6 +440,7 @@ process_option_identity(krb5_context context,
         break;
 #endif
     case IDTYPE_DIR:
+        free(idopts->cert_filename);
         idopts->cert_filename = strdup(residual);
         if (idopts->cert_filename == NULL)
             retval = ENOMEM;
@@ -514,7 +516,7 @@ pkinit_identity_initialize(krb5_context context,
                            krb5_principal princ)
 {
     krb5_error_code retval = EINVAL;
-    int i;
+    size_t i;
 
     pkiDebug("%s: %p %p %p\n", __FUNCTION__, context, idopts, id_cryptoctx);
     if (!(princ &&
@@ -576,9 +578,10 @@ pkinit_identity_prompt(krb5_context context,
                        int do_matching,
                        krb5_principal princ)
 {
-    krb5_error_code retval = EINVAL;
+    krb5_error_code retval = 0;
     const char *signer_identity;
-    int i;
+    krb5_boolean valid;
+    size_t i;
 
     pkiDebug("%s: %p %p %p\n", __FUNCTION__, context, idopts, id_cryptoctx);
     if (!(princ &&
@@ -630,22 +633,36 @@ pkinit_identity_prompt(krb5_context context,
             goto errout;
     } /* Not anonymous principal */
 
+    /* Require at least one successful anchor if any are specified. */
+    valid = FALSE;
     for (i = 0; idopts->anchors != NULL && idopts->anchors[i] != NULL; i++) {
         retval = process_option_ca_crl(context, plg_cryptoctx, req_cryptoctx,
                                        idopts, id_cryptoctx,
                                        idopts->anchors[i], CATYPE_ANCHORS);
-        if (retval)
-            goto errout;
+        if (!retval)
+            valid = TRUE;
     }
+    if (retval && !valid)
+        goto errout;
+    krb5_clear_error_message(context);
+    retval = 0;
+
+    /* Require at least one successful intermediate if any are specified. */
+    valid = FALSE;
     for (i = 0; idopts->intermediates != NULL
              && idopts->intermediates[i] != NULL; i++) {
         retval = process_option_ca_crl(context, plg_cryptoctx, req_cryptoctx,
                                        idopts, id_cryptoctx,
                                        idopts->intermediates[i],
                                        CATYPE_INTERMEDIATES);
-        if (retval)
-            goto errout;
+        if (!retval)
+            valid = TRUE;
     }
+    if (retval && !valid)
+        goto errout;
+    krb5_clear_error_message(context);
+    retval = 0;
+
     for (i = 0; idopts->crls != NULL && idopts->crls[i] != NULL; i++) {
         retval = process_option_ca_crl(context, plg_cryptoctx, req_cryptoctx,
                                        idopts, id_cryptoctx, idopts->crls[i],
@@ -668,7 +685,7 @@ pkinit_set_deferred_id(pkinit_deferred_id **identities,
                        const char *identity, unsigned long ck_flags,
                        const char *password)
 {
-    int i;
+    size_t i;
     pkinit_deferred_id *out = NULL, *ids;
     char *tmp;
 
@@ -730,7 +747,7 @@ const char *
 pkinit_find_deferred_id(pkinit_deferred_id *identities,
                         const char *identity)
 {
-    int i;
+    size_t i;
 
     for (i = 0; identities != NULL && identities[i] != NULL; i++) {
         if (strcmp(identities[i]->identity, identity) == 0)
@@ -747,7 +764,7 @@ unsigned long
 pkinit_get_deferred_id_flags(pkinit_deferred_id *identities,
                              const char *identity)
 {
-    int i;
+    size_t i;
 
     for (i = 0; identities != NULL && identities[i] != NULL; i++) {
         if (strcmp(identities[i]->identity, identity) == 0)
@@ -762,7 +779,7 @@ pkinit_get_deferred_id_flags(pkinit_deferred_id *identities,
 void
 pkinit_free_deferred_ids(pkinit_deferred_id *identities)
 {
-    int i;
+    size_t i;
 
     for (i = 0; identities != NULL && identities[i] != NULL; i++) {
         free(identities[i]->identity);
